﻿using Padsurface.App.Services.Interfaces;
using System;
using System.ComponentModel;
using System.Composition;
using System.Runtime.CompilerServices;
using System.Windows.Input;

namespace Padsurface.App.ViewModels
{
    /// <summary>
    /// Implementation of <see cref="INotifyPropertyChanged"/> to simplify models.
    /// </summary>
    [Windows.Foundation.Metadata.WebHostHidden]
    public abstract class BindableBase : INotifyPropertyChanged
    {
        /// <summary>
        /// Multicast event for property change notifications.
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        [Import]
        public INavigationService NavigationService { get; set; }

        public BindableBase()
        {
            NavigateCommand = new DelegateCommand(Navigate);
        }

        /// <summary>
        /// Checks if a property already matches a desired value.  Sets the property and
        /// notifies listeners only when necessary.
        /// </summary>
        /// <typeparam name="T">Type of the property.</typeparam>
        /// <param name="storage">Reference to a property with both getter and setter.</param>
        /// <param name="value">Desired value for the property.</param>
        /// <param name="propertyName">Name of the property used to notify listeners.  This
        /// value is optional and can be provided automatically when invoked from compilers that
        /// support CallerMemberName.</param>
        /// <returns>True if the value was changed, false if the existing value matched the
        /// desired value.</returns>
        protected bool SetProperty<T>(ref T storage, T value, [CallerMemberName] String propertyName = null)
        {
            if (object.Equals(storage, value)) return false;

            storage = value;
            this.OnPropertyChanged(propertyName);
            return true;
        }

        /// <summary>
        /// Notifies listeners that a property value has changed.
        /// </summary>
        /// <param name="propertyName">Name of the property used to notify listeners.  This
        /// value is optional and can be provided automatically when invoked from compilers
        /// that support <see cref="CallerMemberNameAttribute"/>.</param>
        protected void OnPropertyChanged([CallerMemberName] string propertyName = null)
        {
            var eventHandler = this.PropertyChanged;
            if (eventHandler != null)
            {
                eventHandler(this, new PropertyChangedEventArgs(propertyName));
            }
        }

        #region NavigateCommand

        public ICommand NavigateCommand { get; set; }

        protected virtual void Navigate(object obj)
        {
            if (obj == null) throw new ArgumentNullException("obj");

            var navigationTarget = Type.GetType(obj.ToString());
            if (navigationTarget == null)
                throw new InvalidOperationException(string.Format("Failed to navigate to '{0}', unknown type", obj));

            NavigationService.Navigate(navigationTarget);
        }

        #endregion
    }
}
