﻿using Padsurface.CoreAudio.Enumerations;
using Padsurface.CoreAudio.MultiMediaStructures;
using System;
using System.Runtime.InteropServices;

namespace Padsurface.CoreAudio.Components.WASAPI.Interfaces
{
    /// <summary>
    /// The IAudioClient interface enables a client to create and initialize an audio stream between an audio application and the audio engine (for a shared-mode stream) 
    /// or the hardware buffer of an audio endpoint device (for an exclusive-mode stream). 
    /// 
    /// Note: the interface member functions are declared in the order that the methods appear in the COM interface, this is required for COM interop
    /// </summary>
    [Guid(ComIIDs.IAudioClientIID), InterfaceType(ComInterfaceType.InterfaceIsIUnknown)]
    public interface IAudioClient
    {
        /// <summary>
        /// Initializes the audio stream
        /// </summary>
        /// <param name="shareMode">The sharing mode for the connection. 
        ///     Through this parameter, the client tells the audio engine whether it wants to share the audio endpoint device with other clients. 
        ///     The client should set this parameter to one of the following: AUDCLNT_SHAREMODE_EXCLUSIVE, AUDCLNT_SHAREMODE_SHARED</param>
        /// <param name="streamFlags">Flags to control creation of the stream. 
        ///     The client should set this parameter to 0 or to the bitwise OR of one or more of the AUDCLNT_STREAMFLAGS_XXX Constants or the AUDCLNT_SESSIONFLAGS_XXX Constants.</param>
        /// <param name="bufferDuration">The buffer capacity as a time value. 
        ///     This parameter is of type REFERENCE_TIME and is expressed in 100-nanosecond units. 
        ///     This parameter contains the buffer size that the caller requests for the buffer that the audio application will share with the audio engine (in shared mode) 
        ///     or with the endpoint device (in exclusive mode). If the call succeeds, the method allocates a buffer that is a least this large.</param>
        /// <param name="periodicity">The device period. 
        ///     This parameter can be nonzero only in exclusive mode. 
        ///     In shared mode, always set this parameter to 0. 
        ///     In exclusive mode, this parameter specifies the requested scheduling period for successive buffer accesses by the audio endpoint device. 
        ///     If the requested device period lies outside the range that is set by the device's minimum period and the system's maximum period, then the method clamps the period to that range. 
        ///     If this parameter is 0, the method sets the device period to its default value. 
        ///     To obtain the default device period, call the IAudioClient::GetDevicePeriod method. 
        ///     If the AUDCLNT_STREAMFLAGS_EVENTCALLBACK stream flag is set and AUDCLNT_SHAREMODE_EXCLUSIVE is set as the ShareMode, 
        ///     then hnsPeriodicity must be nonzero and equal to hnsBufferDuration.</param>
        /// <param name="format">Pointer to a format descriptor. This parameter must point to a valid format descriptor of type WAVEFORMATEX (or WAVEFORMATEXTENSIBLE). </param>
        /// <param name="audioSessionGuid">Pointer to a session GUID. 
        ///     This parameter points to a GUID value that identifies the audio session that the stream belongs to. 
        ///     If the GUID identifies a session that has been previously opened, the method adds the stream to that session. 
        ///     If the GUID does not identify an existing session, the method opens a new session and adds the stream to that session. 
        ///     The stream remains a member of the same session for its lifetime. Setting this parameter to NULL is equivalent to passing a pointer to a GUID_NULL value.</param>
        /// <returns>HRESULT: If the method succeeds, it returns S_OK</returns>
        int Initialize(AudioClientShareMode shareMode,
            UInt32 streamFlags,
            long bufferDuration,
            long periodicity,
            [In] WaveFormatEx format,
            [In] ref Guid audioSessionGuid);

        /// <summary>
        /// Retrieves the size (maximum capacity) of the audio buffer associated with the endpoint
        /// </summary>
        /// <param name="numBufferFrames">Pointer to a UINT32 variable into which the method writes the number of audio frames that the buffer can hold</param>
        /// <returns>HRESULT: If the method succeeds, it returns S_OK</returns>
        int GetBufferSize(out UInt32 numBufferFrames);

        /// <summary>
        /// Retrieves the maximum latency for the current stream and can be called any time after the stream has been initialized
        /// </summary>
        /// <param name="latency">Pointer to a REFERENCE_TIME variable into which the method writes a time value representing the latency. 
        ///     The time is expressed in 100-nanosecond units.</param>
        /// <returns>HRESULT: If the method succeeds, it returns S_OK</returns>
        int GetStreamLatency([Out, MarshalAs(UnmanagedType.I8)] out long latency);

        /// <summary>
        /// Retrieves the number of frames of padding in the endpoint buffer
        /// </summary>
        /// <param name="numPaddingFrames">Pointer to a UINT32 variable into which the method writes the frame count (the number of audio frames of padding in the buffer)</param>
        /// <returns>HRESULT: If the method succeeds, it returns S_OK</returns>
        int GetCurrentPadding(out UInt32 numPaddingFrames);

        /// <summary>
        /// Indicates whether the audio endpoint device supports a particular stream format
        /// </summary>
        /// <param name="shareMode">The sharing mode for the stream format. 
        ///     Through this parameter, the client indicates whether it wants to use the specified format in exclusive mode or shared mode. 
        ///     The client should set this parameter to one of the following AUDCLNT_SHAREMODE enumeration values: AUDCLNT_SHAREMODE_EXCLUSIVE, AUDCLNT_SHAREMODE_SHARED
        /// </param>
        /// <param name="format">Pointer to the specified stream format. 
        ///     This parameter points to a caller-allocated format descriptor of type WAVEFORMATEX or WAVEFORMATEXTENSIBLE. 
        ///     The client writes a format description to this structure before calling this method.</param>
        /// <param name="closestMatch">Pointer to a pointer variable into which the method writes the address of a WAVEFORMATEX or WAVEFORMATEXTENSIBLE structure. 
        ///     This structure specifies the supported format that is closest to the format that the client specified through the pFormat parameter. 
        ///     For shared mode (that is, if the ShareMode parameter is AUDCLNT_SHAREMODE_SHARED), set ppClosestMatch to point to a valid, non-NULL pointer variable. 
        ///     For exclusive mode, set ppClosestMatch to NULL. 
        ///     The method allocates the storage for the structure. 
        ///     The caller is responsible for freeing the storage, when it is no longer needed, by calling the CoTaskMemFree function. 
        ///     If the IsFormatSupported call fails and ppClosestMatch is non-NULL, the method sets *ppClosestMatch to NULL.</param>
        /// <returns>HRESULT: If the method succeeds, it returns S_OK</returns>
        int IsFormatSupported(
            AudioClientShareMode shareMode,
            [In] WaveFormatEx format,
            [Out, MarshalAs(UnmanagedType.LPStruct)] out WaveFormatEx closestMatch);

        /// <summary>
        /// Retrieves the stream format that the audio engine uses for its internal processing of shared-mode streams
        /// </summary>
        /// <param name="deviceFormat">Pointer to a pointer variable into which the method writes the address of the mix format. 
        ///     This parameter must be a valid, non-NULL pointer to a pointer variable. 
        ///     The method writes the address of a WAVEFORMATEX (or WAVEFORMATEXTENSIBLE) structure to this variable. 
        ///     The method allocates the storage for the structure. 
        ///     The caller is responsible for freeing the storage, when it is no longer needed, by calling the CoTaskMemFree function. 
        ///     If the GetMixFormat call fails, deviceFormat is NULL.</param>
        /// <returns>HRESULT: If the method succeeds, it returns S_OK</returns>
        int GetMixFormat(out IntPtr deviceFormat);

        /// <summary>
        /// Retrieves the length of the periodic interval separating successive processing passes by the audio engine on the data in the endpoint buffer
        /// </summary>
        /// <param name="defaultDevicePeriod">Pointer to a REFERENCE_TIME variable into which the method writes a time value specifying the default interval 
        ///     between periodic processing passes by the audio engine. The time is expressed in 100-nanosecond units. </param>
        /// <param name="minimumDevicePeriod">Pointer to a REFERENCE_TIME variable into which the method writes a time value specifying the minimum interval 
        ///     between periodic processing passes by the audio endpoint device. The time is expressed in 100-nanosecond units.</param>
        /// <returns>HRESULT: If the method succeeds, it returns S_OK</returns>
        int GetDevicePeriod(out long defaultDevicePeriod, out long minimumDevicePeriod);

        /// <summary>
        /// Starts the audio stream
        /// </summary>
        /// <returns>HRESULT: If the method succeeds, it returns S_OK</returns>
        int Start();

        /// <summary>
        /// Stops the audio stream
        /// </summary>
        /// <returns>HRESULT: If the method succeeds, it returns S_OK</returns>
        int Stop();

        /// <summary>
        /// Resets the audio stream
        /// </summary>
        /// <returns>HRESULT: If the method succeeds, it returns S_OK</returns>
        int Reset();

        /// <summary>
        /// Sets the event handle that the audio engine will signal each time a buffer becomes ready to be processed by the client
        /// </summary>
        /// <returns>HRESULT: If the method succeeds, it returns S_OK</returns>
        int SetEventHandle(IntPtr eventHandle);

        /// <summary>
        /// Accesses additional services from the audio client object
        /// </summary>
        /// <param name="interfaceId">The interface ID for the requested service.
        /// The client should set this parameter to one of the following REFIID values:
        ///  - IID_IAudioCaptureClient
        ///  - IID_IAudioClock
        ///  - IID_IAudioRenderClient
        ///  - IID_IAudioSessionControl
        ///  - IID_IAudioStreamVolume
        ///  - IID_IChannelAudioVolume
        ///  - IID_IMFTrustedOutput
        ///  - IID_ISimpleAudioVolume
        /// </param>
        /// <param name="interfacePointer">Pointer to a pointer variable into which the method writes the address of an instance of the requested interface.
        ///     Through this method, the caller obtains a counted reference to the interface. 
        ///     The caller is responsible for releasing the interface, when it is no longer needed, by calling the interface's Release method. 
        ///     If the GetService call fails, interfacePointer is NULL</param>
        /// <returns>HRESULT: If the method succeeds, it returns S_OK</returns>
        int GetService([In, MarshalAs(UnmanagedType.LPStruct)] Guid interfaceId,
            [Out, MarshalAs(UnmanagedType.IUnknown)] out object interfacePointer);
    }
}